/*---------------------------------------------------------------
 *	PWMを使った音楽再生とLEDイルミネーション(μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2024 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- PWMのレジスタ定義 ------------------------------------
#define PWM_BASE(un)	((un) == 0 ? 0x4001c000 : (un) == 1 ? 0x40021000 : (un) == 2 ? 0x40022000 : 0x4002d000)
#define PWM(un, r)				(PWM_BASE(un) + PWM_##r)
#define PWM_TASKS_STOP			0x004
#define PWM_TASKS_SEQSTART(n)	(0x008 + (n) * 0x04)
#define PWM_LOOPSDONE			0x11C
#define PWM_ENABLE				0x500
#define PWM_MODE				0x504
#define PWM_COUNTERTOP			0x508
#define PWM_PRESCALER			0x50c
#define PWM_DECODER				0x510
#define PWM_LOOP				0x514
#define PWM_SEQ_PTR(n)			(0x520 + (n) * 0x20)
#define PWM_SEQ_CNT(n)			(0x524 + (n) * 0x20)
#define PWM_SEQ_REFRESH(n)		(0x528 + (n) * 0x20)
#define PWM_SEQ_ENDDELAY(n)		(0x52c + (n) * 0x20)
#define PWM_PSEL_OUT(n)			(0x560 + (n) * 0x04)

//-------- LED制御のためのGPIOの初期設定 ------------------------
LOCAL void led_init(void)
{
	// COL1..COL5とROW1..ROW5に接続されたGPIOピンを出力に設定
	out_w(GPIO(P0, PIN_CNF(28)), 1);	// GPIO P0.28(COL1に接続)
	out_w(GPIO(P0, PIN_CNF(11)), 1);	// GPIO P0.11(COL2に接続)
	out_w(GPIO(P0, PIN_CNF(31)), 1);	// GPIO P0.31(COL3に接続)
	out_w(GPIO(P1, PIN_CNF(05)), 1);	// GPIO P1.05(COL4に接続)
	out_w(GPIO(P0, PIN_CNF(30)), 1);	// GPIO P0.30(COL5に接続)
	out_w(GPIO(P0, PIN_CNF(21)), 1);	// GPIO P0.21(ROW1に接続)
	out_w(GPIO(P0, PIN_CNF(22)), 1);	// GPIO P0.22(ROW2に接続)
	out_w(GPIO(P0, PIN_CNF(15)), 1);	// GPIO P0.15(ROW3に接続)
	out_w(GPIO(P0, PIN_CNF(24)), 1);	// GPIO P0.24(ROW4に接続)
	out_w(GPIO(P0, PIN_CNF(19)), 1);	// GPIO P0.19(ROW5に接続)
}

//-------- GPIOの特定のピンpinの出力をval(0または1)に設定 -------
//	pinで指定された以外のピンの状態は変化しない
LOCAL void out_gpio_pin(UW port, UW pin, UW val)
{
	INT port_addr;

	if(port == 1){							// P1の場合
		if(val)
			port_addr = GPIO(P1, OUTSET);	// P1でvalが1の場合
		else
			port_addr = GPIO(P1, OUTCLR);	// P1でvalが0の場合

	} else {								// P0の場合
		if(val)
			port_addr = GPIO(P0, OUTSET);	// P0でvalが1の場合
		else
			port_addr = GPIO(P0, OUTCLR);	// P0でvalが0の場合
	}
	out_w(port_addr, (1 << pin));			// 指定のピンに1を出力
}

// -------------------------------------------------------
// ボタンスイッチA,Bに対応するGPIO P0のピン番号
#define GPIO_P0_SW_A	14
#define GPIO_P0_SW_B	23

#define SPEAKER_GPIO_PORT	0	// SPEAKERに接続されたGPIOのポート番号
#define SPEAKER_GPIO_PIN	0	// SPEAKERに接続されたGPIOのピン番号

// -------------------------------------------------------
// GPIOTEのレジスタ定義
#define GPIOTE(r)			(GPIOTE_BASE + GPIOTE_##r)
#define GPIOTE_BASE			0x40006000
#define GPIOTE_EVENTS_IN(n) (0x100 + (n) * 4)
#define GPIOTE_INTENSET		0x304
#define GPIOTE_CONFIG(n)	(0x510 + (n) * 4)

// GPIOTE CONFIG[n]のレジスタの設定値
#define GPIOTE_CONFIG_Event		1
#define GPIOTE_CONFIG_Task		3
#define GPIOTE_CONFIG_LoToHi	1
#define GPIOTE_CONFIG_HiToLo	2
#define GPIOTE_CONFIG_Toggle	3

// GPIOTEの割込み番号と割込み優先度レベル
#define INTNO_GPIOTE			6	// GPIOTEの割込み番号
#define INTPRI_GPIOTE			6	// GPIOTEの割込み優先度レベル

// -------------------------------------------------------
// LED用PWMのカウンタ上限値の定義(周期時間は1マイクロ秒)
#define	TOP				1000

const INT led_pwm = 0;				 	// unit=0のPWMを使ってLEDを点灯

// -------------------------------------------------------
// LED用PWMのカウンタ比較値を格納する3次元配列
//	comp_table[パターンptn][ステップstep][チャンネルch]
//	&comp_table[ptn][0][0] のアドレスをSEQ_PTR(0,1)に設定
/*
  --------------------------------------------------------------
								ch=0	ch=1	ch=2	ch=3
								 ↓		 ↓		 ↓		 ↓
  --------------------------------------------------------------
	ptn=0	step=0				COMP0	COMP1	COMP2	COMP3
			step=1				COMP0	COMP1	COMP2	COMP3
			...					...		...		...		...
			step=(MAX_STEP-1)	COMP0	COMP1	COMP2	COMP3
  --------------------------------------------------------------
	ptn=1	step=0				COMP0	COMP1	COMP2	COMP3
			step=1				COMP0	COMP1	COMP2	COMP3
			...					...		...		...		...
			step=(MAX_STEP-1)	COMP0	COMP1	COMP2	COMP3
  --------------------------------------------------------------
	ptn=2
								...
  --------------------------------------------------------------
								...
  --------------------------------------------------------------
	ptn=(MAX_PTN-1)
			step=0				COMP0	COMP1	COMP2	COMP3
			step=1				COMP0	COMP1	COMP2	COMP3
			...					...		...		...		...
			step=(MAX_STEP-1)	COMP0	COMP1	COMP2	COMP3
  --------------------------------------------------------------
*/
#define	MAX_PTN			20
#define	PTN_01_OFFSET	10			// SEQ0用ptnとSEQ1用ptnのインデックスの差
#define	PTN_DOREMI_CNT	8			// 音階に対応して使用するパターン数
#define	PTN_SEQ0_MIN	0									// SEQ0用ptnの最小値
#define	PTN_SEQ0_MAX	(PTN_SEQ0_MIN + PTN_DOREMI_CNT - 1)	// SEQ0用ptnの最大値
#define	PTN_SEQ1_MIN	(PTN_SEQ0_MIN + PTN_01_OFFSET)		// SEQ1用ptnの最小値
#define	PTN_SEQ1_MAX	(PTN_SEQ1_MIN + PTN_DOREMI_CNT - 1)	// SEQ1用ptnの最大値
#define	PTN_SEQ0_OFF	(PTN_SEQ0_MAX + 1)		// SEQ0用の点灯OFFパターン
#define	PTN_SEQ1_OFF	(PTN_SEQ1_MAX + 1)		// SEQ1用の点灯OFFパターン

#define	MAX_STEP	16
#define	MAX_CH		4
#define	SEQ0_PERCENT	10		// (SEQ0+SEQ1)に対するSEQ0の時間の割合(%)

// カウンタ比較値を入れる配列
static _UH comp_table[MAX_PTN][MAX_STEP][MAX_CH];

// 千分率のdutyからカウンタ比較値への変換
LOCAL UH duty2comp(INT duty){
	return(TOP - (duty * TOP / 1000));
}

// 指定のパターン、チャンネル、最大デューティ比(max_duty)と減衰率(decay)から、
//	各ステップの比較値をcomp_tableに設定
//	max_dutyとdecayは千分率で指定
//
//	(dir > 0 の場合)
//	comp_table[][step=0][] に max_duty を設定
//	comp_table[][step=1][] に max_duty * (decay / 1000) を設定
//	comp_table[][step=2][] に max_duty * (decay / 1000)^2 を設定
//	comp_table[][step=3][] に max_duty * (decay / 1000)^3 を設定
//			...
//
//	(dir < 0 の場合)
//	comp_table[][step=MAX_STEP-1][] に max_duty を設定
//	comp_table[][step=MAX_STEP-2][] に max_duty * (decay / 1000) を設定
//	comp_table[][step=MAX_STEP-3][] に max_duty * (decay / 1000)^2 を設定
//	comp_table[][step=MAX_STEP-4][] に max_duty * (decay / 1000)^3 を設定
//			...
//
LOCAL void init_step_duty(INT max_duty, INT ptn, INT step_decay,
							INT ch, INT dir){
	for(INT idx = 0; idx < MAX_STEP; idx++){
		comp_table[ptn][dir > 0 ? idx : (MAX_STEP-1-idx)][ch]
			= duty2comp(max_duty);
		max_duty = max_duty * step_decay / 1000;
	}
}

// 距離(差異)に応じて千分率による減衰率を返す
LOCAL INT ch_decay(INT dist){

	// 負数の場合もあるので2乗してから比較
	switch (dist * dist){
		case (0*0):	return(1000);
		case (1*1):	return(300);
		case (2*2):	return(100);
		case (3*3):	return(50);
		case (4*4):	return(10);
		case (5*5):	return(5);
		case (6*6):	return(1);
		case (7*7):	return(0);
		default:
			tm_printf("ch_decay: illegal dist=%d\n", dist);
			return(0);
	}
}

// ドレミの音階へのインデックスdoremiに応じて指定のパターンを生成
//	チャンネル番号(0..3)の2倍がちょうど音階へのインデックス(=0..7)に
//	なるチャンネルでは、max_dutyをそのまま使う
//	その前後のチャンネルでは、両者の距離(差異)に応じてch_decayによりdutyが減衰
LOCAL void init_ptn_duty(INT max_duty, INT ptn,
							INT step_decay, INT doremi, INT dir){
	for(INT idx = 0; idx < MAX_CH; idx++){
		init_step_duty(ch_decay(idx * 2 - doremi) * max_duty / 1000,
			ptn, step_decay, idx, dir);
	}
}

// すべてのパターン番号に対するcomp_tableを設定
LOCAL void init_pwm_duty(){
	INT idx_ptn;

	// ドレミの音階に応じたSEQ0用のパターンの生成
	for(idx_ptn = PTN_SEQ0_MIN; idx_ptn <= PTN_SEQ0_MAX; idx_ptn++){
		init_ptn_duty(1000, idx_ptn, 400, idx_ptn, (-1));
	}

	// ドレミの音階に応じたSEQ1用のパターンの生成
	for(idx_ptn = PTN_SEQ1_MIN; idx_ptn <= PTN_SEQ1_MAX; idx_ptn++){
		init_ptn_duty(1000, idx_ptn, 700, (idx_ptn - PTN_01_OFFSET), 1);
	}

	// すべて消灯するパターン(PTN_SEQ0_OFF,PTN_SEQ1_OFF)の生成
	init_ptn_duty(0, PTN_SEQ0_OFF, 0, 0, 1);
	init_ptn_duty(0, PTN_SEQ1_OFF, 0, 0, 1);
}

// 比較値のテーブルを設定してからPWMをスタート
LOCAL void pwm_duty_start(INT unit,
	_UH *seq0_table, _UH *seq1_table, INT seq0_refresh, INT seq1_refresh){

	tm_printf("pwm_duty_start: seq0tab=%08x, seq1tab=%08x, seq0ref=%d, seq1ref=%d\n", seq0_table, seq1_table, seq0_refresh, seq1_refresh);

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);
	out_w(PWM(unit, PRESCALER), 4);		// DIV_16でクロック1MHzを指定
	out_w(PWM(unit, COUNTERTOP), TOP);	// カウンタ上限値の設定
	out_w(PWM(unit, DECODER), 2);		// LOAD=individual(single)を指定(※D)

	// シーケンス0の比較値の配列の先頭アドレスseq0_tableを設定
	out_w(PWM(unit, SEQ_PTR(0)), (UW)seq0_table);
	out_w(PWM(unit, SEQ_CNT(0)), MAX_STEP * MAX_CH);
	out_w(PWM(unit, SEQ_REFRESH(0)), seq0_refresh);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	// シーケンス1の比較値の配列の先頭アドレスseq1_tableを設定
	out_w(PWM(unit, SEQ_PTR(1)), (UW)seq1_table);
	out_w(PWM(unit, SEQ_CNT(1)), MAX_STEP * MAX_CH);
	out_w(PWM(unit, SEQ_REFRESH(1)), seq1_refresh);
	out_w(PWM(unit, SEQ_ENDDELAY(1)), 0);

	// 繰り返し1回を指定してPWMのシーケンス0の動作を開始、
	out_w(PWM(unit, LOOP), 1);
	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

// -------------------------------------------------------
//	PWMを使ったスピーカーの制御
// -------------------------------------------------------
// 比較値COMP0を格納する16ビット幅のメモリ領域
static _UH half_cycle_us;

// unitで指定されたPWMに周波数freq(Hz)の発振を設定
LOCAL void pwm_speaker_start(INT unit, INT freq)
{
	// 1周期の波形に対するマイクロ秒単位の時間を計算
	// (＝1MHzのクロックに対するカウント値)
	INT cycle_us = (1000 * 1000) / freq;

	// cycle_usの半分の時間(＝1MHzのクロックに対するカウント値の半分)を計算
	half_cycle_us = cycle_us / 2;

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);

	out_w(PWM(unit, PRESCALER), 4);	// クロック1MHzで利用(Divide by 16)

	out_w(PWM(unit, COUNTERTOP), cycle_us);
	out_w(PWM(unit, LOOP), 0);
	out_w(PWM(unit, DECODER), 0);	// LOAD=commonを指定

	out_w(PWM(unit, SEQ_PTR(0)), (UW)&half_cycle_us);
	out_w(PWM(unit, SEQ_CNT(0)), 1);
	out_w(PWM(unit, SEQ_REFRESH(0)), 0);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

// unitで指定されたPWMの発振を停止
LOCAL void pwm_stop(INT unit)
{
	out_w(PWM(unit, TASKS_STOP), 1);
}

// -------------------------------------------------------
const INT speaker_pwm = 1;		// スピーカー用PWMとしてunit=1を使用

// ドレミファソラシド(C5,D5,E5,F5,G5,A5,B5)の周波数(Hz)
const INT pwm_freqs[] = { 523, 587, 659, 698, 783, 880, 987 };

LOCAL W cur_time(void)			// ミリ秒単位の現在時刻の下位32ビットを返す
{
	SYSTIM	tim;
	tk_get_tim(&tim);			// 現在時刻を取得
	return(tim.lo);
}

// -------------------------------------------------------
// 楽譜データとその処理
//	MML(Music Macro Language)の仕様を参考にしているが異なる点もある
//
//	C D E F G A B:	ドレミファソラシに対する音記号
//	R:				休符の音記号
//	>:				次の音記号(1つ)を1オクターブ上げる
//	<:				次の音記号(1つ)を1オクターブ下げる
//	1 2 4 8 6		次の音記号(1つ)が 全音符 1/2 1/4(default) 1/8 1/16音符
//	.:				次の音記号(1つ)の音の長さを1.5倍にする
//	space:			無視

INT	pwm_4time;		// 1/4音符に対する音の時間(ms)
INT	pwm_n_time;		// 次の音記号に対する音の時間(ms)
INT	pwm_n_octave;	// 次の音記号に対する1オクターブ上下((-1)/0/1)

INT	prev_time;		// 音符の再生を開始した時刻

LOCAL void init_pwm_play(void)	// 各音符の処理前の初期化
{
	pwm_n_time = pwm_4time;
	pwm_n_octave = 0;
}

// 起動時のチャイム
B* music_start[] = {"4C4E4G2.>C R", NULL};

// かえるの合唱(※M)
B* music_kaeru[] = {
	"4C4D4E4F 4E4D4.C8R ", "4E4F4G4A 4G4F4.E8R ",
	"4C4R4C4R 4C4R4C4R ",  "8C8C8D8D8E8E8F8F 4E4D4.C1R ",
	NULL};

// 春が来た(※N)
B* music_haru[] = {
	"4G8E8F4G4A 4G8E8F4G4>C", "4A4G4.E8C 2.D4R",
	"4G8A8G4E4G 4>C8>D8>C4A4>C", "4G4>E4.>D8G 1>C R",
	NULL};

// 1つの音符の再生(※P)
LOCAL void pwm_play(B m_char)
{
	INT	freq, doremi, ptn;
	INT	seq0_time, seq1_time, seq0_refresh, seq1_refresh;
	_UH	*seq0_table, *seq1_table;

	if(m_char == 'R'){									// 休符の場合
		seq0_table = &comp_table[PTN_SEQ0_OFF][0][0]; 	// LEDを消灯
		seq1_table = &comp_table[PTN_SEQ1_OFF][0][0];
	} else {
		if((m_char < 'A') || (m_char > 'G')){			// 音階の指定
			tm_printf("pwm_play: illegal char=%c(%02x)\n", m_char, m_char);
			return;
		}
		doremi = m_char - 'C';					// 音階のインデックス
		if(doremi < 0) { doremi += 7; }			// 'A','B' の場合
		ptn = doremi;

		freq = pwm_freqs[doremi];				// 音階に対応する周波数を取得
		if(pwm_n_octave > 0){					// オクターブを上げる場合
			freq = freq << pwm_n_octave;		// 1オクターブで周波数2倍
			ptn = 7;							// 一番高い音のLED点灯パターン
		} else if(pwm_n_octave < 0){			// オクターブを下げる場合
			freq = freq >> (- pwm_n_octave);	// 1オクターブで周波数1/2
			ptn = 0;							// 一番低い音のLED点灯パターン
		}

		// スピーカー用PWMを制御して音符を再生(※Q)
		pwm_speaker_start(speaker_pwm, freq);

		// LED用PWMのパラメータとしてptnに応じた点灯パターンを設定(※S)
		seq0_table = &comp_table[PTN_SEQ0_MIN + ptn][0][0];
		seq1_table = &comp_table[PTN_SEQ1_MIN + ptn][0][0];
	}

	seq0_time = pwm_n_time * SEQ0_PERCENT / 100;	// ミリ秒単位のSEQ0時間
	seq1_time = pwm_n_time - seq0_time;				// ミリ秒単位のSEQ1時間

	// SEQ0,SEQ1のミリ秒単位の時間からREFRESHレジスタの設定値を計算
	//	1ステップのサイクル数 = ミリ秒単位の1ステップの所要時間
	//							= (seq0_time / MAX_STEP)
	seq0_refresh =	seq0_time / MAX_STEP - 1;
	if(seq0_refresh < 0) seq0_refresh = 0;		// (-1)になった場合は0に戻す
	seq1_refresh =	seq1_time / MAX_STEP - 1;
	if(seq1_refresh < 0) seq1_refresh = 0;		// (-1)になった場合は0に戻す

	// LED用PWMを制御してLEDを点灯(※R)
	pwm_duty_start(led_pwm, seq0_table, seq1_table, seq0_refresh, seq1_refresh);

	prev_time += pwm_n_time;			// 音符の再生を終了すべき時刻の計算
	tk_dly_tsk(prev_time - cur_time());	// 終了すべき時刻までディレイ

	pwm_stop(speaker_pwm);				// 音符の再生を終了
	tk_dly_tsk(10);						// 各音符の後のディレイ
}

// 音符の長さとオクターブ上下の処理
LOCAL void pwm_ctl_time(B m_char)
{
	if(m_char == '1' || m_char == '2' || m_char == '4' || m_char == '8'){
		pwm_n_time = pwm_4time * 4 / (m_char - '0');	// 1,2,4,8分音符
	} else if(m_char == '6'){
		pwm_n_time = pwm_4time * 4 / 16;				// 16分音符
	} else if(m_char == '.'){
		pwm_n_time = pwm_n_time * 3 / 2;				// 1.5倍の処理
	} else if(m_char == '>'){
		pwm_n_octave += 1;								// 1オクターブ上げる
	} else if(m_char == '<'){
		pwm_n_octave -= 1;								// 1オクターブ下げる
	} else if(m_char != ' '){
		tm_printf("pwm_ctl_time: illegal char=%c(%02x)\n", m_char, m_char);
	}
}

// 連続した音符(＝楽譜)の再生
LOCAL void pwm_music(B** ptr_music, INT tempo){
	pwm_4time = tempo;
	prev_time = cur_time();			// 音符の再生開始時の時刻を記録
	init_pwm_play();				// 音符の再生の前の初期化

	while(*ptr_music != NULL){
		B m_char;
		INT	index = 0;

		// 各音記号の処理
		while((m_char = (*ptr_music)[index++]) != NULL){
			if(m_char >= 'A'){
				pwm_play(m_char);		// 音符の再生
				init_pwm_play();		// 音符の再生の前の初期化
			} else {
				pwm_ctl_time(m_char);	// 音符の長さとオクターブ上下の処理
			}
		}
		ptr_music++;					// 次の音記号に進む
	}
}

// -------------------------------------------------------
//	ボタンスイッチによる割込み発生と音楽の再生
// -------------------------------------------------------
// オブジェクトID番号
ID tskid;			// メッセージ表示用タスクのID
ID flgid;			// 割込み通知用イベントフラグのID

// ボタンスイッチ入力のためのGPIOの初期設定
LOCAL void btn_init(void)
{
	// ボタンスイッチに対応するGPIO P0のピンを入力に設定
	out_w(GPIO(P0, PIN_CNF(GPIO_P0_SW_A)), 0);
	out_w(GPIO(P0, PIN_CNF(GPIO_P0_SW_B)), 0);
}

// GPIOTEの割込みハンドラ
LOCAL void btn_inthdr(UINT intno)
{
	INT n;						// GPIOTEのチャンネル番号(0または1)
	for (n = 0; n < 2; n++) {	// GPIOTEのチャンネル0とチャンネル1を順に処理

		// EVENTS_IN[n]の最下位ビットを見てイベント発生状態を確認
		if (in_w(GPIOTE(EVENTS_IN(n))) & (1 << 0)) {

			// イベントがあった場合はイベント発生状態をクリア
			out_w(GPIOTE(EVENTS_IN(n)), 0);

			// 割込み通知用イベントフラグの対応ビットをセット
			tk_set_flg(flgid, (1 << n));
		}
	}
}

// -------------------------------------------------------
// ボタンA、Bが押されたら音楽を再生するタスク(※T)
// -------------------------------------------------------
void play_task(INT stacd, void *exinf)
{
	BOOL btn_a, btn_b;
	UINT flgptn;

	pwm_music(music_start, 80);

	for(;;){									// 永久に繰り返し

		// 割込み通知用イベントフラグの下位2ビットをOR待ち
		tk_wai_flg(flgid, 0b11, (TWF_ORW | TWF_CLR), &flgptn, TMO_FEVR);

		if((flgptn & (1 << 0)) != 0){			// 最下位ビットが1の場合
			tm_printf("Button_SW_A: pushed\n"); // ボタンスイッチAの押下を表示
			pwm_music(music_kaeru, 400);		// 「かえるの合唱」を再生
		}
		if((flgptn & (1 << 1)) != 0){		// 最下位から2番目のビットが1の場合
			tm_printf("Button_SW_B: pushed\n"); // ボタンスイッチBの押下を表示
			pwm_music(music_haru, 500);			// 「春が来た」を再生
		}

		// 500ミリ秒間待つ、その間の再度のボタン押下はチャタリングとして無視
		tk_dly_tsk(500);
		tk_clr_flg(flgid, 0);
	}
}

// タスクとイベントフラグの生成情報
const T_CTSK ctsk = {0, (TA_HLNG | TA_RNG3), &play_task, 10, 1024, 0};
const T_CFLG cflg = {0, (TA_TFIFO | TA_WSGL), 0};

// 割込みハンドラ定義情報
const T_DINT dint = {TA_HLNG, btn_inthdr};

// -------------------------------------------------------
//	メインプログラム
// -------------------------------------------------------
EXPORT void usermain( void )
{
	// SPEAKERに接続されたGPIOピン(P0.00)を出力に設定
	out_w(GPIO(P0, PIN_CNF(SPEAKER_GPIO_PIN)), (1 << 0));

	// PWM(unit=0)のチャンネル0をGPIOピン(P0.00)経由でSPEAKERに出力
	out_w(PWM(speaker_pwm, PSEL_OUT(0)),
			((SPEAKER_GPIO_PORT << 5) | (SPEAKER_GPIO_PIN << 0)));

	led_init();

	// LEDを点灯させるためのGPIOの設定(※F)
	out_gpio_pin(0, 28, 0);		// GPIO P0.28-COL1に0を出力
	out_gpio_pin(0, 11, 0);		// GPIO P0.11-COL2に0を出力
	out_gpio_pin(0, 31, 0);		// GPIO P0.31-COL3に0を出力
	out_gpio_pin(1,  5, 0);		// GPIO P1.05-COL4に0を出力
	out_gpio_pin(0, 30, 0);		// GPIO P0.30-COL5に0を出力

    // PWMの各チャンネル出力をGPIOピン経由でROW5..ROW2に出力(※A)
	out_w(PWM(led_pwm, PSEL_OUT(0)), ((0 << 5) | (19 << 0)));
										// チャンネル0→P0.19(ROW5に接続)
	out_w(PWM(led_pwm, PSEL_OUT(1)), ((0 << 5) | (24 << 0)));
										// チャンネル1→P0.24(ROW4に接続)
	out_w(PWM(led_pwm, PSEL_OUT(2)), ((0 << 5) | (15 << 0)));
										// チャンネル2→P0.15(ROW3に接続)
	out_w(PWM(led_pwm, PSEL_OUT(3)), ((0 << 5) | (22 << 0)));
										// チャンネル3→P0.22(ROW2に接続)

	btn_init();				// ボタンスイッチ入力のためのGPIOの初期設定

	// 以下、ボタンスイッチの押下でGPIOTEのイベントを発生するための設定
	// ボタンスイッチAでチャンネル0のイベントを発生
	out_w(GPIOTE(CONFIG(0)),				// チャンネル0を指定
		  (GPIOTE_CONFIG_Event << 0)		// MODEにEventを指定
		  | (GPIO_P0_SW_A << 8)				// PSELにSW_AのGPIOピン番号14を指定
		  | (0 << 13)						// PORTにSW_AのGPIOポート番号0を指定
		  | (GPIOTE_CONFIG_HiToLo << 16));	// POLARITYにHiToLo(立ち下がり)を指定

	// ボタンスイッチBでチャンネル1のイベントを発生
	out_w(GPIOTE(CONFIG(1)),				// チャンネル1を指定
		  (GPIOTE_CONFIG_Event << 0)		// MODEにEventを指定
		  | (GPIO_P0_SW_B << 8)				// PSELにSW_BのGPIOピン番号23を指定
		  | (0 << 13)						// PORTにSW_BのGPIOポート番号0を指定
		  | (GPIOTE_CONFIG_HiToLo << 16));	// POLARITYにHiToLo(立ち下がり)を指定

	// GPIOTEのチャンネル0とチャンネル1の割込みを許可
	out_w(GPIOTE(INTENSET), (1 << 0) | (1 << 1));

	init_pwm_duty();						// カウンタ比較値の配列データを設定

	flgid = tk_cre_flg(&cflg);				// 割込み通知用のイベントフラグ作成
	tskid = tk_cre_tsk(&ctsk);				// 音楽再生用のタスク生成
	tk_sta_tsk(tskid, 0);					// 音楽再生用のタスク起動

	tk_def_int(INTNO_GPIOTE, &dint);		// ボタン入力用の割込みハンドラ定義
	EnableInt(INTNO_GPIOTE, INTPRI_GPIOTE); // 割込み許可

	tk_slp_tsk(TMO_FEVR);					// 永久待ち
}
